/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "xml/xml.pmh"

#define FE_SAX_DEBUG		FALSE

namespace fe
{

BWORD SurfaceAccessibleXml::load(String a_filename,
	sp<StringFilterI> a_spFilter)
{
#if FE_SAX_DEBUG
	feLog("SurfaceAccessibleXml::load \"%s\"\n",a_filename.c_str());
#endif

	pugi::xml_document document;
	pugi::xml_parse_result result=document.load_file(a_filename.c_str());
	if(!result)
	{
		return result;
	}

	//* TODO param
	const Real triScale=1.0;

	const Color red(1.0,0.0,0.0);
	const Color green(0.0,1.0,0.0);

	sp<SurfaceAccessorI> spOutputPoint=accessor(e_point,e_position,
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputColor=accessor(e_point,"Cd",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputNormal=accessor(e_primitive,"N",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputName=accessor(e_primitive,"name",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputParent=accessor(e_primitive,"parent",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputRefX=accessor(e_primitive,"refX",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputRefY=accessor(e_primitive,"refY",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputRefZ=accessor(e_primitive,"refZ",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputRefT=accessor(e_primitive,"refT",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputAnimX=accessor(e_primitive,"animX",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputAnimY=accessor(e_primitive,"animY",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputAnimZ=accessor(e_primitive,"animZ",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputAnimT=accessor(e_primitive,"animT",
			SurfaceAccessibleI::e_createMissing);
	sp<SurfaceAccessorI> spOutputVertices=accessor(e_primitive,e_vertices,
			SurfaceAccessibleI::e_createMissing);

	std::map<String,SpatialTransform> hiddenTransform;
	std::map<String,String> delegateName;
	std::map<String,SpatialTransform> refMap;
	std::map<String,SpatialTransform> animMap;
	std::map<String,SpatialTransform> refTweakMap;
	std::map<String,SpatialTransform> animTweakMap;
	std::map<String,SpatialTransform> deltaMap;

#if FE_SAX_DEBUG
	feLog("SurfaceAccessibleXml::load result \"%s\"\n",result.description());
#endif

	//* TODO check type and version
	const pugi::xml_node root=document.first_child();

	const pugi::xml_node collection=root.first_child();

	for(pugi::xml_node child=collection.first_child();child;
			child=child.next_sibling())
	{
		const String name=interpretString(child,"name");

		const String parentName=interpretString(child,"parent");

		const SpatialTransform refWorld=
				interpretTransform(child,"worldreference");
		const SpatialVector relativeTranslation=
				interpretVector(child,"translation");
		const SpatialTransform relativeRotation=
				interpretEuler(child,"rotation","rotationorder");

		SpatialTransform relativeMove;
		setIdentity(relativeMove);
		translate(relativeMove,relativeTranslation);

		SpatialTransform xformParent;
		SpatialTransform xformHidden;
		String delegate=parentName;

		const BWORD hasParent=(animMap.find(parentName)!=animMap.end());
		if(hasParent)
		{
			delegate=delegateName[parentName];
			xformHidden=hiddenTransform[parentName];
			xformParent=animMap[parentName];
		}
		else
		{
			setIdentity(xformParent);
			setIdentity(xformHidden);
		}

		const SpatialTransform xformRelative=relativeRotation*relativeMove;
		const SpatialTransform xformWorld=xformRelative*xformParent;

		refMap[name]=refWorld;
		animMap[name]=xformWorld;

		BWORD skip=FALSE;

		if(a_spFilter.isValid() && a_spFilter->filter(name).empty())
		{
			skip=TRUE;
		}

		if(skip)
		{
			hiddenTransform[name]=xformRelative*xformHidden;
			delegateName[name]=delegate;
			refTweakMap[name]=refWorld;
			animTweakMap[name]=xformWorld;
			setIdentity(deltaMap[name]);
			continue;
		}

		SpatialTransform xformParentRefTweak;
		SpatialTransform xformParentTweak;
		SpatialTransform deltaParent;

		if(hasParent)
		{
			xformParentRefTweak=refTweakMap[parentName];
			xformParentTweak=animTweakMap[parentName];
			deltaParent=deltaMap[parentName];
		}
		else
		{
			setIdentity(xformParentRefTweak);
			setIdentity(xformParentTweak);
			setIdentity(deltaParent);
		}

		setIdentity(hiddenTransform[name]);
		delegateName[name]=name;
		SpatialTransform invRefWorld;
		invert(invRefWorld,refWorld);
		const SpatialTransform delta=invRefWorld*xformWorld;

		const SpatialQuaternion quat=xformRelative;

		//* find local tweak in reference

		SpatialTransform invParentRefTweak;
		invert(invParentRefTweak,xformParentRefTweak);

		SpatialTransform refTweak=xformParentRefTweak;

		const SpatialVector toWorld=
				refWorld.translation()-xformParentRefTweak.translation();
		const SpatialVector toLocal=rotateVector(invParentRefTweak,toWorld);
		if(magnitudeSquared(toLocal)>1e-6)
		{
			const SpatialVector fromLocal(0.0,1.0,0.0);

			SpatialQuaternion quatLocal;
			set(quatLocal,fromLocal,unitSafe(toLocal));

			const SpatialTransform localTweak=quatLocal;

			refTweak=localTweak*refTweak;
		}
		setTranslation(refTweak,refWorld.translation());

		SpatialTransform animTweak=refTweak*deltaParent;
		setTranslation(animTweak,xformWorld.translation());

		refTweakMap[name]=refTweak;
		animTweakMap[name]=animTweak;
		deltaMap[name]=delta;

		const SpatialVector color=interpretVector(child,"color");

		const I32 primitiveIndex=spOutputVertices->append(3);
		spOutputName->set(primitiveIndex,name);
		spOutputParent->set(primitiveIndex,parentName);
		spOutputNormal->set(primitiveIndex,animTweak.column(1));

		const I32 pointIndex0=spOutputVertices->integer(primitiveIndex,0);
		const I32 pointIndex1=spOutputVertices->integer(primitiveIndex,1);
		const I32 pointIndex2=spOutputVertices->integer(primitiveIndex,2);

		const SpatialVector point=animTweak.translation();
		spOutputPoint->set(pointIndex0,point);
		spOutputPoint->set(pointIndex1,point+triScale*animTweak.column(0));
		spOutputPoint->set(pointIndex2,point+1.5*triScale*animTweak.column(2));

		spOutputColor->set(pointIndex0,color);
		spOutputColor->set(pointIndex1,red);
		spOutputColor->set(pointIndex2,green);

		//* ported

		SpatialTransform invParentTweak;
		invert(invParentTweak,xformParentTweak);

		SpatialTransform xformLocalRef=refTweak*invParentRefTweak;
		SpatialTransform xformLocal=animTweak*invParentTweak;

		spOutputRefX->set(primitiveIndex,xformLocalRef.column(0));
		spOutputRefY->set(primitiveIndex,xformLocalRef.column(1));
		spOutputRefZ->set(primitiveIndex,xformLocalRef.column(2));
		spOutputRefT->set(primitiveIndex,xformLocalRef.translation());

		spOutputAnimX->set(primitiveIndex,xformLocal.column(0));
		spOutputAnimY->set(primitiveIndex,xformLocal.column(1));
		spOutputAnimZ->set(primitiveIndex,xformLocal.column(2));
		spOutputAnimT->set(primitiveIndex,xformLocal.translation());

#if FE_SAX_DEBUG
		const SpatialVector degrees=interpretVector(child,"rotation");
		const SpatialVector scale=interpretVector(child,"scale");

		feLog("%s \"%s\" parent \"%s\"\n",
				child.name(),name.c_str(),parentName.c_str());
		feLog("  T %s R %s\n  S %s Off %s\n",
				c_print(relativeTranslation),c_print(degrees),
				c_print(scale),c_print(offset));
		feLog("  rot\n%s\n",c_print(relativeRotation));
		feLog("  xformWorld\n%s\n",c_print(xformWorld));
		feLog("  ref\n%s\n",c_print(refWorld));
		feLog("  delta\n%s\n",c_print(refWorld));
		feLog("  refTweak\n%s\n",c_print(refTweak));
		feLog("  animTweak\n%s\n",c_print(animTweak));
		feLog("  prim %d points %d %d %d\n\n",
				primitiveIndex,pointIndex0,pointIndex1,pointIndex2);
#endif
	}

	return result;
}

SpatialTransform SurfaceAccessibleXml::interpretEuler(
	pugi::xml_node& a_rNode,String a_attrName,String a_attrOrder)
{
	const SpatialVector degrees=interpretVector(a_rNode,a_attrName);
	const String order=interpretString(a_rNode,a_attrOrder);

	const SpatialEuler euler(
			degrees[0]*degToRad,degrees[1]*degToRad,degrees[2]*degToRad);

	if(order=="ZYX")
	{
		return euler;
	}

	//* TODO move to Euler<>

	SpatialTransform xform;
	setIdentity(xform);

	if(order=="ZXY")
	{
		rotate(xform,euler[2],e_zAxis);
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[1],e_yAxis);
	}
	else if(order=="XZY")
	{
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[2],e_zAxis);
		rotate(xform,euler[1],e_yAxis);
	}
	else if(order=="YZX")
	{
		rotate(xform,euler[1],e_yAxis);
		rotate(xform,euler[2],e_zAxis);
		rotate(xform,euler[0],e_xAxis);
	}
	else if(order=="XYZ")
	{
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[1],e_yAxis);
		rotate(xform,euler[2],e_zAxis);
	}
	else if(order=="YXZ")
	{
		rotate(xform,euler[1],e_yAxis);
		rotate(xform,euler[0],e_xAxis);
		rotate(xform,euler[2],e_zAxis);
	}

	return xform;
}

String SurfaceAccessibleXml::interpretString(
	pugi::xml_node& a_rNode,String a_attrName)
{
	pugi::xml_attribute xmlAttr=a_rNode.attribute(a_attrName.c_str());

	return xmlAttr.value();
}

SpatialVector SurfaceAccessibleXml::interpretVector(
	pugi::xml_node& a_rNode,String a_attrName)
{
	pugi::xml_attribute xmlAttr=a_rNode.attribute(a_attrName.c_str());

	float x=0.0;
	float y=0.0;
	float z=0.0;
	sscanf(xmlAttr.value(),"%f,%f,%f",&x,&y,&z);

	return SpatialVector(x,y,z);
}

SpatialTransform SurfaceAccessibleXml::interpretTransform(
	pugi::xml_node& a_rNode,String a_attrName)
{
	pugi::xml_attribute xmlAttr=a_rNode.attribute(a_attrName.c_str());

	float v[16];
	sscanf(xmlAttr.value(),
			"%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f",
			&v[0],&v[1],&v[2],&v[3],
			&v[4],&v[5],&v[6],&v[7],
			&v[8],&v[9],&v[10],&v[11],
			&v[12],&v[13],&v[14],&v[15]);

	SpatialTransform matrix;
	set(matrix,v);

	return matrix;
}

} // namespace
